from ipaddress import ip_address
import json
import logging
from tornado.web import RequestHandler, urlparse, HTTPError
from tornado.escape import utf8
from werkzeug.http import parse_options_header

from .. import exc
from .. import models
from ..settings import settings


# Logging object
log = logging.getLogger(__name__)


class BaseHandler(RequestHandler):
    def initialize(self):
        self.session = self.application.my_settings.get("db_session")()

    def on_finish(self):
        self.session.close()

    def get_or_create_user(self, email):
        """Get a user or create a new one and return a User object."""
        user = self.session.query(models.User).filter_by(email=email).first()
        if not user:
            user = models.User(email=email).add(self.session)
            self.session.commit()

        return user

    def get_current_user(self):
        """Default global user fetch by user_auth_header."""

        # Fetch the email address from the auth_header (e.g. X-NSoT-Email)
        auth_header = settings.user_auth_header
        log.debug('  fetching auth_header: %s' % auth_header)
        email = self.request.headers.get(auth_header)

        if email is not None:
            log.debug('auth_header authenticated user: %s' % email)
            user = self.get_or_create_user(email)
            return user
        return None

    def get_client_ip(self):
        address, port = self.request.connection.context.address
        if not isinstance(address, unicode):
            address = unicode(address)
        return ip_address(address)

    def _is_permitted_network(self):
        if not settings.restrict_networks:
            return True

        ip = self.get_client_ip()
        for network in settings.restrict_networks:
            if ip in network:
                return True

        return False

    def prepare(self):
        log.debug('BaseHandler.prepare()')
        if not self._is_permitted_network():
            raise exc.Forbidden("Connected from forbidden network.")

        try:
            if not self.current_user:
                raise exc.Unauthorized("Not logged in.")
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)


class FeHandler(BaseHandler):

    def prepare(self):
        BaseHandler.prepare(self)
        # Need to access token to set Cookie.
        self.xsrf_token

    def render_template(self, template_name, **kwargs):
        template = self.application.my_settings["template_env"].get_template(
            template_name
        )
        content = template.render(kwargs)
        return content

    def render(self, template_name, **kwargs):
        context = {}
        context.update(self.get_template_namespace())
        context.update(kwargs)
        self.write(self.render_template(template_name, **context))

    def write_error(self, status_code, **kwargs):
        message = "An unknown problem has occured :("
        if "exc_info" in kwargs:
            inst = kwargs["exc_info"][1]
            if isinstance(inst, HTTPError):
                message = inst.log_message
            else:
                message = str(inst)

        # Pass context to the error template
        self.render("error.html", code=status_code, message=message)


class ApiHandler(BaseHandler):
    def initialize(self):
        BaseHandler.initialize(self)
        self._jbody = None

    @property
    def jbody(self):
        if self._jbody is None:
            if self.request.body:
                self._jbody = json.loads(self.request.body)
            else:
                self._jbody = {}
        return self._jbody

    def extract_token_credentials(self):
        """Extract email/auth_token from request"""
        authz = self.request.headers.get('Authorization')
        if authz and authz.lower().startswith('authtoken '):
            log.debug('Getting email/auth_token from Authorization header.')
            auth_type, data = authz.split()
            email, auth_token = data.split(':', 1)
        else:
            log.debug('Getting email/auth_token from arguments.')
            email = self.get_argument('email', None)
            auth_token = self.get_argument('auth_token', None)
        log.debug('     email: %s' % email)
        log.debug('auth_token: %s' % auth_token)
        return email, auth_token

    def get_current_user(self):
        """Try user_auth_header, then try auth_token header."""
        # Perform default authentication (user_auth_header)
        user = super(ApiHandler, self).get_current_user()
        if user is not None:
            log.debug('API login default method succeeded!')
            return user

        # Args used for auth_token auth...
        email, auth_token = self.extract_token_credentials()

        # Do we want these here or just to return a generice 401 "Login
        # required"?
        if email is None:
            raise exc.Unauthorized('Missing Required argument: email')
        if auth_token is None:
            raise exc.Unauthorized('Missing Required argument: auth_token')

        user = models.User.verify_auth_token(email, auth_token)

        # If user is bad this time, it's an invalid login
        if user is None:
            raise exc.Unauthorized('Invalid login/token expired.')

        log.debug('token_auth authenticated user: %s' % email)
        return user

    def check_xsrf_cookie(self):
        """Optionally check XSRF cookies on API calls."""
        if settings.api_xsrf_enabled:
            super(ApiHandler, self).check_xsrf_cookie()

    def get_pagination_values(self, max_limit=None):
        offset = int(self.get_argument("offset", 0))
        limit = self.get_argument("limit", None)
        if limit is None:
            return offset, limit

        limit = int(limit)
        if max_limit is not None and limit > max_limit:
            limit = max_limit

        return offset, limit

    def paginate_query(self, query, offset, limit):
        total = query.count()

        query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)

        return query, total

    def prepare(self):
        BaseHandler.prepare(self)

        if self.request.method.lower() in ("put", "post"):
            content_type = parse_options_header(
                self.request.headers.get("Content-Type")
            )[0]
            if content_type.lower() != "application/json":
                raise exc.BadRequest("Invalid Content-Type for POST/PUT request.")

    def write_error(self, status_code, **kwargs):

        message = "An unknown problem has occured :("
        if "exc_info" in kwargs:
            inst = kwargs["exc_info"][1]
            if isinstance(inst, HTTPError):
                message = inst.log_message
            else:
                message = str(inst)

        self.write({
            "status": "error",
            "error": {
                "code": status_code,
                "message": message,
            },
        })

    def success(self, data):
        """200 OK"""
        self.write({
            "status": "ok",
            "data": data,
        })
        self.finish()

    def created(self, location=None, data=None):
        """201 CREATED"""
        self.set_status(201)
        if data is None:
            data = {}
        if location is not None:
            self.set_header(
                "Location",
                urlparse.urljoin(utf8(self.request.uri), utf8(location))
            )
        self.write({
            "status": "ok",
            "data": data,
        })
        self.finish()



# -*- coding: utf-8 -*-

"""
do some crypto
"""

# Asymmetric key crypto
"""
from cryptography.fernet import Fernet

key = Fernet.generate_key()
f = Fernet(key)
token = f.encrypt('my deep dark secret')
print 'crypted token =', token
print f.decrypt(token)
"""


import hashlib
from hashlib import sha1
import json
import requests
import time


host = "https://tails.corp.dropbox.com"
client = "pynsot-dev"
clientVersion = "0.1"
clientDescription = 'test cilent for pynsot'
user = 'jathan'
get_authToken = lambda: str(int(time.time()))

cert = "n2phaq35gihfob3mfbtgyzpg4lgnf7xhuouilndk7ff3gvkzoaceorxdhgf5hc3uio2kvfidocgxrhz6pliee6bfl5nb3tiuilc3i25turx62ubnxpcngk7efd7gzgv2ilkz5v5h23b5qrcmzbysewj7jtstkrlw7xmdgdpco4ettatcnmaxw6whurtervl5bwotqqhoyoezdxduq37ph5erbkdo5qlm6o2rs7vu4w25hm25d6hfix2h5ckpcg7"

authToken = get_authToken()
def get_authSignature(authToken, cert):
    key = authToken + cert
    sig = sha1(key)
    return sig.hexdigest()

authSignature = get_authSignature(authToken, cert)
uri = '/api/conduit.connect'
url = host + uri

params = dict(
    client=client,
    clientVersion=clientVersion,
    clientDescription=clientDescription,
    user=user,
    host=host,
    authToken=authToken,
    authSignature=authSignature,
)

data = dict(
    params=json.dumps(params),
    output='json',
    __conduit__=True,
)

# Make the request to conduit.connect
r = requests.post(url, data=data)

# Parse the response (no error-handling yet)
result = r.json()['result']
conduit = {
    'sessionKey': result['sessionKey'],
    'connectionID': result['connectionID'],
}

# Good response:
"""
{u'error_code': None,
 u'error_info': None,
 u'result': {u'connectionID': 562985,
  u'sessionKey': u'wjnlt6wnpvcx2uajclrb3i5n4o3x3yzspim2uf7r',
  u'userPHID': u'PHID-USER-ugykvugfbmv4ijqlh5ln'}}
"""

# Bad response:
"""
{u'error_code': u'ERR-INVALID-USER',
 u'error_info': u'The username you are attempting to authenticate with is not valid.',
 u'result': None}
"""



import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app, user, session, auth_token
from .util import (
    assert_error, assert_success,
)



def test_no_user(tornado_server):
    assert_error(requests.get(
        "http://localhost:{}/api/sites".format(tornado_server.port)
    ), 401)


def test_valid_user(tornado_server):
    assert_success(requests.get(
        "http://localhost:{}/api/sites".format(tornado_server.port),
        headers={"X-NSoT-Email": "gary@localhost"}
    ), {"sites": [], "limit": None, "offset": 0, "total": 0})


def test_invalid_user(tornado_server):
    assert_error(requests.get(
        "http://localhost:{}/api/sites".format(tornado_server.port),
        headers={"X-NSoT-Email": "gary"}
    ), 400)


def test_get_auth_token_valid(tornado_server, user):
    """Test that an auth_token can be generated."""
    payload = {"email": user.email, "secret_key": user.secret_key}
    data = json.dumps(payload)
    r = requests.post(
        "http://localhost:{}/api/authenticate".format(tornado_server.port),
        headers={"Content-Type": "application/json"},
        data=data,
    )
    # Since assert_success is checking the payload, we're retreiving and
    # comparing it first.
    response_data = r.json()['data']
    assert_success(r, response_data)


def test_get_auth_token_invalid(tornado_server, user):
    """Test that an auth_token fails w/ a bad secret key."""
    payload = {"email": user.email, "secret_key": "bogus"}
    data = json.dumps(payload)
    assert_error(requests.post(
        "http://localhost:{}/api/authenticate".format(tornado_server.port),
        headers={"Content-Type": "application/json"},
        data=data,
    ), 401)


def test_verify_auth_token_valid(tornado_server, user, auth_token):
    """Test that an auth_token is valid."""
    assert_success(requests.post(
        "http://localhost:{}/api/verify_token".format(tornado_server.port),
        headers={
            "Content-Type": "application/json",
            "Authorization": "AuthToken {}:{}".format(user.email, auth_token)
        },
    ), True)


def test_verify_auth_token_invalid(tornado_server, user):
    """Test that an auth_token is NOT valid."""
    assert_error(requests.post(
        "http://localhost:{}/api/verify_token".format(tornado_server.port),
        headers={
            "Content-Type": "application/json",
            "Authorization": "AuthToken {}:{}".format(user.email, 'bogus')
        },
    ), 401)


def test_valid_auth_token(tornado_server, user, auth_token):
    """Test that a GET can be performed to a resource using auth_token."""
    assert_success(requests.get(
        "http://localhost:{}/api/sites".format(tornado_server.port),
        headers={
            "Content-Type": "application/json",
            "Authorization": "AuthToken {}:{}".format(user.email, auth_token)
        },
    ), {"sites": [], "limit": None, "offset": 0, "total": 0})



from .handlers import api, fe

HANDLERS = [

    # Sites
    (r"/api/sites", api.SitesHandler),
    (r"/api/sites/(?P<site_id>\d+)", api.SiteHandler),

    # Attributes
    (r"/api/sites/(?P<site_id>\d+)/attributes", api.AttributesHandler),
    (r"/api/sites/(?P<site_id>\d+)/attributes/(?P<attribute_id>\d+)", api.AttributeHandler),

    # Devices
    (r"/api/sites/(?P<site_id>\d+)/devices", api.DevicesHandler),
    (r"/api/sites/(?P<site_id>\d+)/devices/query", api.DevicesQueryHandler),
    (r"/api/sites/(?P<site_id>\d+)/devices/(?P<device_id>\d+)", api.DeviceHandler),

    # Networks
    (r"/api/sites/(?P<site_id>\d+)/networks", api.NetworksHandler),
    (r"/api/sites/(?P<site_id>\d+)/networks/query", api.NetworksQueryHandler),
    (r"/api/sites/(?P<site_id>\d+)/networks/(?P<network_id>\d+)", api.NetworkHandler),
    (   r"/api/sites/(?P<site_id>\d+)/networks/(?P<network_id>\d+)/subnets",
        api.NetworkSubnetsHandler
    ),
    (   r"/api/sites/(?P<site_id>\d+)/networks/(?P<network_id>\d+)/supernets",
        api.NetworkSupernetsHandler
    ),

    # Change Log
    (r"/api/sites/(?P<site_id>\d+)/changes", api.ChangesHandler),
    (r"/api/sites/(?P<site_id>\d+)/changes/(?P<change_id>\d+)", api.ChangeHandler),

    # Users
    (r"/api/users", api.UsersHandler),
    (r"/api/users/(?P<user_id>\d+)", api.UserHandler),
    (r"/api/users/(?P<user_id>\d+)/rotate_secret_key", api.UserRotateSecretKeyHandler),
    (r"/api/users/(?P<user_id>\d+)/permissions", api.UserPermissionsHandler),
    (
        r"/api/users/(?P<user_id>\d+)/permissions/(?P<site_id>\d+)",
        api.UserPermissionHandler
    ),

    # API Auth
    (r"/api/authenticate", api.AuthTokenLoginHandler),
    (r"/api/verify_token", api.AuthTokenVerifyHandler),

    # API fallback
    (r"/api/.*", api.NotFoundHandler),

    # Frontend Handlers
    (r".*", fe.AppHandler),
]



import yaml

from ipaddress import ip_network


class Settings(object):
    def __init__(self, initial_settings):
        self.settings = initial_settings

    @classmethod
    def from_settings(cls, settings, initial_settings=None):
        _settings = {}
        _settings.update(settings.settings)
        if initial_settings:
            _settings.update(initial_settings)
        return cls(_settings)

    def update_from_config(self, filename):
        with open(filename) as config:
            data = yaml.safe_load(config.read())

        settings = {}
        settings.update(data)

        for key, value in settings.iteritems():
            key = key.lower()

            if key not in self.settings:
                continue

            override = getattr(self, "override_%s" % key, None)
            if override is not None and callable(override):
                value = override(value)

            self.settings[key] = value

    def __getitem__(self, key):
        return self.settings[key]

    def __getattr__(self, name):
        try:
            return self.settings[name]
        except KeyError as err:
            raise AttributeError(err)

    # Overrides

    def override_restrict_networks(self, values):
        if not isinstance(values, (list, set)):
            values = [values]

        new_values = []
        for value in values:
            if not isinstance(value, unicode):
                value = unicode(value)
            new_values.append(ip_network(value))

        return new_values


settings = Settings({
    "log_format": "%(asctime)-15s\t%(levelname)s\t%(message)s",
    "num_processes": 1,
    "database": None,
    "debug": False,
    "port": 8990,
    "user_auth_header": "X-NSoT-Email",
    "restrict_networks": [],
    "bind_address": None,
    "api_xsrf_enabled": True,
    "secret_key": "SECRET_KEY",
    "auth_token_expiry": 600,  # 10 minutes
})



import bittle

# These values cannot be reordered. Only add new
# permissions to the end.
PermissionsFlag = bittle.FlagWord([
    "admin",        # Someone who is an admin has full permissions
                    # within a site. Admins are able to grant other
                    # permissions as well as Create/Update/Delete all
                    # resources.
])



import json
import os
import requests
import urllib
import urlparse


def assert_error(response, code):
    output = response.json()
    assert output["status"] == "error"
    assert output["error"]["code"] == code


def assert_success(response, data=None):
    output = response.json()
    assert response.status_code == 200
    assert output["status"] == "ok"
    if data is not None:
        print 'OUTPUT DATA = %r' % (output['data'],)
        print 'QUERY DATA = %r' % (data,)
        assert output["data"] == data


def assert_created(response, location, data=None):
    output = response.json()
    assert response.status_code == 201
    assert response.headers.get("Location") == location
    assert output["status"] == "ok"
    if data is not None:
        assert output["data"] == data


def assert_deleted(response):
    output = response.json()
    assert response.status_code == 200
    assert output["status"] == "ok"


class Client(object):
    def __init__(self, tornado_server, user="user"):
        self.tornado_server = tornado_server
        self.user = "{}@localhost".format(user)

    @property
    def base_url(self):
        return "http://localhost:{}/api".format(self.tornado_server.port)

    def request(self, method, url, user="admin", **kwargs):

        headers = {
            "X-NSoT-Email": self.user
        }

        if method.lower() in ("put", "post"):
            headers["Content-type"] = "application/json"

        return requests.request(
            method, self.base_url + url,
            headers=headers, **kwargs
        )

    def get(self, url, **kwargs):
        return self.request("GET", url, **kwargs)

    def post(self, url, **kwargs):
        return self.request("POST", url, **kwargs)

    def put(self, url, **kwargs):
        return self.request("PUT", url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request("DELETE", url, **kwargs)

    def create(self, url, **kwargs):
        return self.post(url, data=json.dumps(kwargs))

    def update(self, url, **kwargs):
        return self.put(url, data=json.dumps(kwargs))


def load_json(relpath):
    """
    Load JSON files relative to this directory.

    Files are loaded from the 'data' directory. So for example for
    ``/path/to/data/devices/foo.json`` the ``relpath`` would be
    ``devices/foo.json``.

    :param relpath:
        Relative path to our directory's "data" dir
    """
    our_path = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(our_path, 'data')
    filepath = os.path.join(data_dir, relpath)
    with open(filepath, 'rb') as f:
        return json.load(f)


def run_set_queries(resource_name, client, device_queries):
    """
    Run set queries on the specified resource.

    The directory structure is expected to match the resource path. So
    "devices/query" would map to both the data directory for JSON response
    files, and the API URL of "/api/sites/1/devices/query".

    The ``device_queries is expected to be a list of 2-tuples of (query,
    filename) where query is a set query and filename is a file containing the
    expected response JSON. Example::

        [
            ('foo=bar', 'test1.json'),
            ('bar=baz', 'test2.json'),
        ]

    :param resource_name:
        The resource for which to run query tests

    :param client:
        A ``Client`` instance

    :param devices_queries:
        A list of 2-tuples of (query, filename)
    """
    base_path = '/sites/1'
    path = os.path.join(resource_name, 'query')
    base_uri = os.path.join(base_path, path)
    uri = base_uri + '?query='  # '/sites/1/devices/query?query='

    # Walk the query and filename, construct a URL w/ the query, load up the
    # file, call the URL w/ the client, and compare the return data to the
    # loaded data.
    for query, filename in device_queries:
        qs = urllib.quote_plus(query)  # 'foo=bar' => 'foo%3Dbar'
        url = uri + qs  # => /sites/1/devices/query?query=foo%3Dbar
        filepath = os.path.join(path, filename)  # devices/query/query1.json
        output = load_json(filepath)
        assert_success(
            client.get(url),
            output['data']
        )



"""
Project-wide utilities.
"""

import collections
import logging
import tornado
from jinja2 import Environment, PackageLoader

from .settings import settings


log = logging.getLogger(__name__)

_TRUTHY = set([
    "true", "yes", "1", ""
])


def qp_to_bool(arg):
    return str(arg).lower() in _TRUTHY


def get_loglevel(args):
    verbose = args.verbose * 10
    quiet = args.quiet * 10
    return logging.getLogger().level - verbose + quiet


def get_template_env(package="nsot", extra_filters=None, extra_globals=None):
    filters = {}
    j_globals = {}

    if extra_filters:
        filters.update(extra_filters)

    if extra_globals:
        j_globals.update(extra_globals)

    env = Environment(
        loader=PackageLoader(package, "templates"),
        extensions=['jinja2.ext.autoescape'],
        autoescape=True

    )
    env.filters.update(filters)
    env.globals.update(j_globals)

    return env


#: Namedtuple for resultant items from ``parse_set_query()``
SetQuery = collections.namedtuple('SetQuery', 'action name value')


def parse_set_query(query):
    """
    Parse a representation of set operations for attribute/value pairs into
    (action, name, value) and return a list of ``SetQuery`` objects.

    Computes left-to-right evaluation, where the first character indicates the
    set operation:

    + "+" indicates a union
    + "-" indicates a difference
    + no marker indicates an intersection

    For example::

        >>> parse_set_query('+owner=team-networking')
        [SetQuery(action='union', name='owner', value='team-networking')]
        >>> parse_set_query('foo=bar')
        [SetQuery(action='intersection', name='foo', value='bar')]
        >>> parse_set_query('foo=bar -owner=team-networking')
        [SetQuery(action='intersection', name='foo', value='bar'),
         SetQuery(action='difference', name='owner', value='team-networking')]

    :param query:
        Set query string
    """
    log.debug('Incoming query = %r' % (query,))
    queries = query.split()

    attributes = []
    for q in queries:
        if q.startswith('+'):
            action = 'union'
            q = q[1:]
        elif q.startswith('-'):
            action = 'difference'
            q = q[1:]
        else:
            action = 'intersection'

        name, _, value = q.partition('=')
        attributes.append(SetQuery(action, name, value))

    log.debug('Outgoing attributes = %r' % (attributes,))
    return attributes




import tornado
import logging
import mimetypes

from .routes import HANDLERS

class Application(tornado.web.Application):
    def __init__(self, *args, **kwargs):

        mimetypes.add_type("application/font-woff2", ".woff2")

        kwargs["handlers"] = HANDLERS
        self.my_settings = kwargs.pop("my_settings", {})
        super(Application, self).__init__(*args, **kwargs)



"""
Test NSoT utilities.
"""

from nsot.util import SetQuery, parse_set_query


def test_parse_set_query():
    """
    Parse a bunch of set queries and make sure that the expected result
    matches.
    """
    # List of 2-tuples of (query, expected_result)
    set_tests = {
        'foo=bar': [
            ('intersection', 'foo', 'bar'),
        ],
        'foo=bar owner=jathan': [
            ('intersection', 'foo', 'bar'),
            ('intersection', 'owner', 'jathan'),
        ],
        '-owner=gary': [
            ('difference', 'owner', 'gary'),
        ],
        'cluster +foo=baz': [
            ('intersection', 'cluster', ''),
            ('union', 'foo', 'baz'),
        ],
        # Extra white space
        'cluster=lax   +foo=baz': [
            ('intersection', 'cluster', 'lax'),
            ('union', 'foo', 'baz'),
        ],
    }

    # Make sure that result matches expected_result
    for query, expected_result in set_tests.iteritems():
        result = parse_set_query(query)
        assert result == expected_result



import base64
import json
import logging
import os
import pytest
import requests
import socket
import threading
import tornado
import tornado.httpserver
import tornado.ioloop
from tornado import netutil

from nsot import models
from nsot.settings import settings
from nsot.app import Application


sa_log = logging.getLogger("sqlalchemy.engine.base.Engine")

# Uncomment to have all queries printed out
# sa_log.setLevel(logging.INFO)

# Overload the server's secret_key with a randomly-generated one.
SECRET_KEY = base64.urlsafe_b64encode(os.urandom(32))
settings.settings['secret_key'] = SECRET_KEY


class Server(object):
    """ Wrapper around Tornado server with test helpers. """

    def __init__(self, tornado_app):
        self.server = tornado.httpserver.HTTPServer(
            tornado_app
        )
        self.server.add_sockets(netutil.bind_sockets(
            None, "localhost", family=socket.AF_INET
        ))
        self.server.start()
        self.io_thread = threading.Thread(
            target=tornado.ioloop.IOLoop.instance().start
        )
        self.io_thread.start()

    @property
    def port(self):
        return self.server._sockets.values()[0].getsockname()[1]


@pytest.fixture()
def tornado_app(request, tmpdir):
    db_path = tmpdir.join("nsot.sqlite")
    db_engine = models.get_db_engine("sqlite:///%s" % db_path)

    models.Model.metadata.drop_all(db_engine)
    models.Model.metadata.create_all(db_engine)
    models.Session.configure(bind=db_engine)

    my_settings = {
        "db_engine": db_engine,
        "db_session": models.Session,
    }

    tornado_settings = {
        "debug": False,
        "cookie_secret": "dropbox",
        "secret_key": SECRET_KEY,  # Setting secret_key here does nothing?!
    }

    return Application(my_settings=my_settings, **tornado_settings)


@pytest.fixture()
def tornado_server(request, tornado_app):

    server = Server(tornado_app)

    def fin():
        tornado.ioloop.IOLoop.instance().stop()
        server.io_thread.join()
    request.addfinalizer(fin)

    return server


## Forklifted from tests/model_tests/fixtures.py.
##
@pytest.fixture
def session(request, tmpdir):
    db_path = tmpdir.join("nsot.sqlite")
    db_engine = models.get_db_engine("sqlite:///%s" % db_path)

    models.Model.metadata.drop_all(db_engine)
    models.Model.metadata.create_all(db_engine)
    models.Session.configure(bind=db_engine)
    session = models.Session()

    def fin():
        session.close()
    request.addfinalizer(fin)

    return session


@pytest.fixture
def user(session):
    """Create a user that we can work with."""
    user = models.User(email="gary@localhost").add(session)
    session.commit()
    return user
##
## END Forklifting


@pytest.fixture
def auth_token(tornado_server, user):
    """Token used to test resources requiring auth_token authentication."""
    payload = {"email": user.email, "secret_key": user.secret_key}
    data = json.dumps(payload)
    r = requests.post(
        "http://localhost:{}/api/authenticate".format(tornado_server.port),
        headers={"Content-Type": "application/json"},
        data=data,
    )
    return r.json()['data']['auth_token']



import pytest
from sqlalchemy.exc import IntegrityError

from nsot import exc
from nsot import models

from .fixtures import session, site, user, admin


def test_creation(session, site, admin):
    models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="test_attribute"
    )
    session.commit()

    attributes = session.query(models.Attribute).all()
    assert len(attributes) == 1
    assert attributes[0].id == 1
    assert attributes[0].site_id == site.id
    assert attributes[0].name == "test_attribute"
    assert attributes[0].required == False


def test_conflict(session, site, admin):
    models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="test_attribute"
    )

    with pytest.raises(IntegrityError):
        models.Attribute.create(
            session, admin.id,
            resource_name="Network",
            site_id=site.id, name="test_attribute"
        )

    models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="test_attribute_2"
    )


def test_validation(session, site, admin):
    with pytest.raises(exc.ValidationError):
        models.Attribute.create(
            session, admin.id,
            resource_name="Network",
            site_id=site.id, name=None,
        )

    with pytest.raises(exc.ValidationError):
        models.Attribute.create(
            session, admin.id,
            resource_name="Network",
            site_id=site.id, name="",
        )

    attribute = models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="test_attribute"
    )

    with pytest.raises(exc.ValidationError):
        attribute.update(admin.id, name="")

    with pytest.raises(exc.ValidationError):
        attribute.update(admin.id, name=None)

    attribute.update(admin.id, name="test_attribute_new")
    session.commit()


def test_deletion(session, site, admin):
    attribute = models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="test_attribute"
    )

    network = models.Network.create(
        session, admin.id, site.id,
        cidr=u"10.0.0.0/8", attributes={"test_attribute": "foo"}
    )

    with pytest.raises(IntegrityError):
        attribute.delete(admin.id)

    network.delete(admin.id)
    attribute.delete(admin.id)

def test_required(session, site, admin):
    attribute_1 = models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="required_1", required=True
    )

    attribute_2 = models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="required_2", required=True
    )

    with pytest.raises(exc.ValidationError):
        network = models.Network.create(
            session, admin.id, site.id,
            cidr=u"10.0.0.0/8"
        )

    with pytest.raises(exc.ValidationError):
        network = models.Network.create(
            session, admin.id, site.id,
            cidr=u"10.0.0.0/8", attributes={
                "required_1": "foo",
            }
        )

    network = models.Network.create(
        session, admin.id, site.id,
        cidr=u"10.0.0.0/8", attributes={
            "required_1": "foo",
            "required_2": "bar",
        }
    )


def test_multi(session, site, admin):
    multi = models.Attribute.create(
        session, admin.id,
        resource_name="Network", display=True,
        site_id=site.id, name="multi", multi=True
    )

    not_multi = models.Attribute.create(
        session, admin.id,
        resource_name="Network",
        site_id=site.id, name="not_multi", multi=False
    )

    models.Network.create(session, admin.id, site.id, cidr=u"10.0.0.0/8")

    models.Network.create(
        session, admin.id, site.id, cidr=u"10.0.0.1",
        attributes={
            "multi": ["test", "testing", "testtttt"]
        }
    )

    with pytest.raises(exc.ValidationError):
        models.Network.create(
            session, admin.id, site.id, cidr=u"10.0.0.2",
            attributes={
                "not_multi": ["test", "testing", "testtttt"]
            }
        )


def test_constraints(session, site, admin):
    default = models.Attribute.create(
        session, admin.id,
        resource_name="Network", site_id=site.id, name="default"
    )

    allow_empty = models.Attribute.create(
        session, admin.id,
        resource_name="Network", site_id=site.id, name="allow_empty",
        constraints={"allow_empty": True}
    )

    pattern = models.Attribute.create(
        session, admin.id,
        resource_name="Network", site_id=site.id, name="pattern",
        constraints={"pattern": "\d\d\d+"}
    )

    valid = models.Attribute.create(
        session, admin.id,
        resource_name="Network", site_id=site.id, name="valid",
        constraints={"valid_values": ["foo", "bar", "baz"]}
    )

    network = models.Network.create(session, admin.id, site.id, cidr=u"10.0.0.0/8")

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"default": ""})

    # Test allow_empty
    network.update(admin.id, attributes={"allow_empty": ""})

    # Test pattern
    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"pattern": ""})

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"pattern": "foo"})

    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"pattern": "10"})

    network.update(admin.id, attributes={"pattern": "100"})
    network.update(admin.id, attributes={"pattern": "1000000"})

    # Test valid_values
    with pytest.raises(exc.ValidationError):
        network.update(admin.id, attributes={"valid": "hello"})

    network.update(admin.id, attributes={"valid": "foo"})



import pytest
from sqlalchemy.exc import IntegrityError

from nsot import exc
from nsot import models

from .fixtures import session, user


def test_site_creation(session, user):
    models.Site.create(
        session, user.id,
        name="Test Site",
        description="This is a Test Site."
    )
    session.commit()

    sites = session.query(models.Site).all()
    assert len(sites) == 1
    assert sites[0].id == 1
    assert sites[0].name == "Test Site"
    assert sites[0].description == "This is a Test Site."


def test_site_conflict(session, user):
    models.Site.create(
        session, user.id,
        name="Test Site",
        description="This is a Test Site."
    )

    with pytest.raises(IntegrityError):
        models.Site.create(
            session, user.id,
            name="Test Site",
            description="This is a Test Site."
        )

    models.Site.create(
        session, user.id,
        name="Test Site 2",
        description="This is a Test Site."
    )


def test_site_validation(session, user):
    with pytest.raises(exc.ValidationError):
        models.Site.create(
            session, user.id,
            name=None,
            description="This is a Test Site."
        )

    with pytest.raises(exc.ValidationError):
        models.Site.create(
            session, user.id,
            name="",
            description="This is a Test Site."
        )

    site = models.Site.create(
        session, user.id,
        name="Test Site",
        description="This is a Test Site."
    )

    with pytest.raises(exc.ValidationError):
        site.update(user.id, name="")

    with pytest.raises(exc.ValidationError):
        site.update(user.id, name=None)

    site.update(user.id, name="Test Site New")
    session.commit()



from tornado.web import HTTPError

class Error(Exception):
    """ Baseclass for NSoT Exceptions."""

class ModelError(Error):
    """ Baseclass for NSoT Model Exceptions."""

class ValidationError(ModelError):
    """ Raised when validation fails on a model."""

class BaseHttpError(HTTPError):
    def __init__(self, log_message, *args, **kwargs):
        HTTPError.__init__(
            self, self.status_code, log_message, *args, **kwargs
        )

class BadRequest(BaseHttpError): status_code = 400
class Unauthorized(BaseHttpError): status_code = 401
class Forbidden(BaseHttpError): status_code = 403
class NotFound(BaseHttpError): status_code = 404
class Conflict(BaseHttpError): status_code = 409



import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client
)


def test_creation(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")  # 1

    # Test invalid attribute name
    assert_error(
        client.create(
            "/sites/1/attributes",
            resource_name="Network", name="invalid attr1"
        ),
        400
    )

    # Successfully create an attribute
    assert_created(
        client.create(
            "/sites/1/attributes",
            resource_name="Network", name="attr1"
        ),
        "/api/sites/1/attributes/1"
    )

    # Successfully get all Network Attributes
    assert_success(
        client.get("/sites/1/attributes"),
        {
            "attributes": [
                {
                    "id": 1, "name": "attr1", "description": "",
                    "required": False, "resource_name": "Network",
                    "site_id": 1, "display": False, "multi": False,
                    "constraints": {
                        "allow_empty": False,
                        "pattern": "",
                        "valid_values": []
                    }
                },
            ],
            "limit": None,
            "offset": 0,
            "total": 1,
        }

    )

    # Successfully get a single Network Attribute
    assert_success(
        client.get("/sites/1/attributes/1"),
        {"attribute": {
            "id": 1, "name": "attr1", "description": "", "resource_name": "Network",
            "required": False, "site_id": 1, "display": False, "multi": False,
            "constraints": {
                "allow_empty": False,
                "pattern": "",
                "valid_values": []
            }
        }}

    )


def test_collection_creation(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")  # 1

    # Successfully create a collection of Attributes
    collection = [
        {"name": "attr1", "resource_name": "Network"},
        {"name": "attr2", "resource_name": "Network"},
        {"name": "attr3", "resource_name": "Network"},
    ]
    collection_response = client.create(
        "/sites/1/attributes",
        attributes=collection
    )
    assert_created(collection_response, None)

    # Successfully get all created Attributes
    output = collection_response.json()
    output['data'].update({"limit": None, "offset": 0})

    assert_success(
        client.get("/sites/1/attributes"),
        output['data'],
    )


def test_update(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")  # 1
    client.create(
        "/sites/1/attributes",
        resource_name="Network", name="attr1"
    )  # 1

    assert_success(
        client.update("/sites/1/attributes/1", description="Attribute 1"),
        {"attribute": {
            "id": 1, "name": "attr1", "description": "Attribute 1",
            "required": False, "site_id": 1, "resource_name": "Network",
            "display": False, "multi": False,
            "constraints": {
                "allow_empty": False,
                "pattern": "",
                "valid_values": []
            }
        }}
    )

    assert_success(
        client.update("/sites/1/attributes/1", required=True),
        {"attribute": {
            "id": 1, "name": "attr1", "description": "",
            "required": True, "site_id": 1, "resource_name": "Network",
            "display": True, "multi": False,
            "constraints": {
                "allow_empty": False,
                "pattern": "",
                "valid_values": []
            }
        }}
    )

    assert_success(
        client.update("/sites/1/attributes/1"),
        {"attribute": {
            "id": 1, "name": "attr1", "description": "",
            "required": False, "site_id": 1, "resource_name": "Network",
            "display": False, "multi": False,
            "constraints": {
                "allow_empty": False,
                "pattern": "",
                "valid_values": []
            }
        }}
    )


def test_deletion(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")
    client.create(
        "/sites/1/attributes",
        resource_name="Network", name="attr1"
    )
    client.create("/sites/1/networks",
        cidr="10.0.0.0/24", attributes={"attr1": "foo"}
    )

    # Don't allow delete when there's an attached network
    assert_error(client.delete("/sites/1/attributes/1"), 409)

    client.delete("/sites/1/networks/1")

    assert_deleted(client.delete("/sites/1/attributes/1"))




import json
import pytest
import requests

from .fixtures import tornado_server, tornado_app
from .util import (
    assert_error, assert_success, assert_created, assert_deleted, Client,
    load_json, run_set_queries
)


def test_creation(tornado_server):
    admin_client = Client(tornado_server, "admin")
    user_client = Client(tornado_server, "user")

    admin_client.create("/sites", name="Test Site")  # 1
    admin_client.create(
        "/sites/1/attributes",
        resource_name="Network", name="attr1"
    )  # 1

    # Invalid permissions
    assert_error(
        user_client.create("/sites/1/networks",
            cidr="10.0.0.0/24", attributes={"attr1": "foo"}
        ),
        403
    )

    # Missing required field (cidr)
    assert_error(
        admin_client.create("/sites/1/networks",
            attributes={"attr1": "foo"}
        ),
        400
    )

    # Null cidr
    assert_error(
        admin_client.create("/sites/1/networks",
            cidr=None,
        ),
        400
    )

    # Verify Successful Creation
    assert_created(
        admin_client.create("/sites/1/networks",
            cidr="10.0.0.0/24", attributes={"attr1": "foo"}
        ),
        "/api/sites/1/networks/1"
    )

    # Verify Successful get of all Networks
    assert_success(
        admin_client.get("/sites/1/networks"),
        {
            "networks": [{
                "attributes": {"attr1": "foo"},
                "id": 1,
                "parent_id": None,
                "ip_version": "4",
                "is_ip": False,
                "network_address": "10.0.0.0",
                "prefix_length": 24,
                "site_id": 1,
            }],
            "limit": None,
            "offset": 0,
            "total": 1,
        }

    )

    # Verify Successful get of single Network
    assert_success(
        admin_client.get("/sites/1/networks/1"),
        {"network": {
            "attributes": {"attr1": "foo"},
            "id": 1,
            "parent_id": None,
            "ip_version": "4",
            "is_ip": False,
            "network_address": "10.0.0.0",
            "prefix_length": 24,
            "site_id": 1
        }}
    )


def test_set_queries(tornado_server):
    """Test set queries for Networks."""
    client = Client(tornado_server)

    client.create('/sites', name='Test Site')  # 1

    # Pre-load the attributes
    attr_data = load_json('attributes.json')
    client.create(
        '/sites/1/attributes',
        attributes=attr_data['attributes']
    )

    # Populate the device objects.
    network_data = load_json('networks.json')
    client.create(
        '/sites/1/networks',
        networks=network_data['networks']
    )

    # Mapping of query string to file containing expected response data for
    # each query.
    network_queries = (
        # INTERSECTION: foo=bar
        ('foo=bar', 'query1.json'),
        # INTERSECTION: foo=bar owner=jathan
        ('foo=bar owner=jathan', 'query2.json'),
        # DIFFERENCE: -owner=gary
        ('-owner=gary', 'query3.json'),
        # UNION: cluster +foo=baz
        ('cluster +foo=baz', 'query4.json'),
    )
    run_set_queries('networks', client, network_queries)


def test_collection_creation(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")  # 1

    # Successfully create a collection of Networks
    collection = [
        {"cidr": "1.1.1.0/24"},
        {"cidr": "2.2.2.0/24"},
        {"cidr": "3.3.3.0/24"},
    ]
    collection_response = client.create(
        "/sites/1/networks",
        networks=collection
    )
    assert_created(collection_response, None)

    # Successfully get all created Networks
    output = collection_response.json()
    output['data'].update({"limit": None, "offset": 0})

    assert_success(
        client.get("/sites/1/networks"),
        output['data'],
    )


def test_update(tornado_server):
    admin_client = Client(tornado_server, "admin")
    user_client = Client(tornado_server, "user")

    admin_client.create("/sites", name="Test Site")  # 1
    admin_client.create(
        "/sites/1/attributes", resource_name="Network", name="attr1"
    )  # 1
    admin_client.create("/sites/1/networks",
        cidr="10.0.0.0/24", attributes={"attr1": "foo"}
    )

    # Empty Update should only clear attributes.
    assert_success(
        admin_client.update("/sites/1/networks/1"),
        {"network": {
            "attributes": {},
            "id": 1,
            "parent_id": None,
            "ip_version": "4",
            "is_ip": False,
            "network_address": "10.0.0.0",
            "prefix_length": 24,
            "site_id": 1
        }}
    )

    # Now put attributes back
    assert_success(
        admin_client.update("/sites/1/networks/1", attributes={"attr1": "foo"}),
        {"network": {
            "attributes": {"attr1": "foo"},
            "id": 1,
            "parent_id": None,
            "ip_version": "4",
            "is_ip": False,
            "network_address": "10.0.0.0",
            "prefix_length": 24,
            "site_id": 1
        }}
    )

    # Invalid permissions
    assert_error(
        user_client.update("/sites/1/networks/1"),
        403
    )

def test_deletion(tornado_server):
    client = Client(tornado_server)

    client.create("/sites", name="Test Site")

    client.create("/sites/1/networks", cidr="10.0.0.0/24")  # 1
    client.create("/sites/1/networks", cidr="10.0.0.1/32")  # 2

    # Don't allow delete when there's an attached subnet/ip
    assert_error(client.delete("/sites/1/networks/1"), 409)

    client.delete("/sites/1/networks/2")

    assert_deleted(client.delete("/sites/1/networks/1"))




import ipaddress
import pytest
import time

from nsot import exc
from nsot import models

from model_tests.fixtures import session, admin, site


def test_create_1024(session, admin, site):

    uid = admin.id
    sid = site.id

    address = u"10.0.0.0/20"
    models.Network.create(session, uid, sid, address)
    models.Attribute.create(session, uid, resource_name="Network", site_id=sid, name="aaaa")

    start = time.time()
    network = ipaddress.ip_network(address)
    for ip in network.subnets(new_prefix=30):
        models.Network.create(session, uid, sid, ip.exploded, {
            "aaaa": "value",
        }, commit=False)
    session.commit()

    print "Finished in {} seconds.".format(time.time() - start)



import logging
from sqlalchemy import desc
from sqlalchemy.exc import IntegrityError


from .util import ApiHandler
from .. import exc
from ..decorators import any_perm
from .. import models
from ..util import qp_to_bool as qpbool, parse_set_query


log = logging.getLogger(__name__)


class SitesHandler(ApiHandler):

    def post(self):
        """ **Create a Site**

        **Example Request**:

        .. sourcecode:: http

            POST /api/sites HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "name": "New Site",
                "description": "This is our new Site."
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 201 OK
            Location: /api/sites/1

            {
                "status": "ok",
                "data": {
                    "site": {
                        "id": 1,
                        "name": "New Site",
                        "description": "This is our new Site."
                    }
                }
            }

        :reqjson string name: The name of the Site
        :reqjson string description: (*optional*) A helpful description for the Site

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.
        :reqheader X-NSoT-Email: required for all api requests.

        :resheader Location: URL to the created resource.

        :statuscode 201: The site was successfully created.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 409: There was a conflict with another resource.
        """

        try:
            name = self.jbody["name"]
            description = self.jbody.get("description", "")
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))
        except ValueError as err:
            raise exc.BadRequest(err.message)

        try:
            site = models.Site.create(
                self.session, self.current_user.id, name=name, description=description
            )
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)

        self.created("/api/sites/{}".format(site.id), {
            "site": site.to_dict(),
        })

    def get(self):
        """ **Get all Sites**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "sites": [
                        {
                            "id": 1
                            "name": "Site 1",
                            "description": ""
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :reqheader X-NSoT-Email: required for all api requests.

        :query string name: (*optional*) Filter to site with name.
        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        """
        name = self.get_argument("name", None)

        sites = self.session.query(models.Site)
        if name is not None:
            sites = sites.filter_by(name=name)

        offset, limit = self.get_pagination_values()
        sites, total = self.paginate_query(sites, offset, limit)

        self.success({
            "sites": [site.to_dict() for site in sites.all()],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class SiteHandler(ApiHandler):
    def get(self, site_id):
        """ **Get a specific Site**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "site": {
                        "id": 1,
                        "name": "Site 1",
                        "description": ""
                    }
                }
            }

        :param site_id: ID of the Site to retrieve
        :type site_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))
        self.success({
            "site": site.to_dict(),
        })

    @any_perm("admin")
    def put(self, site_id):
        """ **Update a Site**

        **Example Request**:

        .. sourcecode:: http

            PUT /api/sites/1 HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "name": "Old Site",
                "description": "A better description."
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "site": {
                        "id": 1,
                        "name": "Old Site",
                        "description": "A better description."
                    }
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.
        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site at site_id was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        try:
            name = self.jbody["name"]
            description = self.jbody.get("description", "")
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

        try:
            site.update(
                self.current_user.id,
                name=name, description=description
            )
        except IntegrityError as err:
            raise exc.Conflict(str(err.orig))

        self.success({
            "site": site.to_dict(),
        })

    @any_perm("admin")
    def delete(self, site_id):
        """ **Delete a Site**

        **Example Request**:

        .. sourcecode:: http

            DELETE /api/sites/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "message": Site 1 deleted."
                }
            }

        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site at site_id was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        try:
            site.delete(self.current_user.id)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)

        self.success({
            "message": "Site {} deleted.".format(site_id),
        })


class AttributesHandler(ApiHandler):

    @any_perm("admin")
    def post(self, site_id):
        """ **Create an Attribute or collection of Attributes**

        **Example Request**:

        .. sourcecode:: http

            POST /api/sites/1/attributes HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "name": "owner",
                "description": "Owner Attribute.",
                "resource_name": "Network",
                "required": false
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 201 OK
            Location: /api/sites/1/attributes/1

            {
                "status": "ok",
                "data": {
                    "attribute": {
                        "id": 1,
                        "name": "owner",
                        "description": "Owner Attribute.",
                        "resource_name": "Network",
                        "required": false,
                        "display": false,
                        "multi": false,
                        "constraints": {
                            "allow_empty": false,
                            "pattern": "",
                            "valid_values": []
                        }
                    }
                }
            }

        You may also create a collection of objects by submitting them as a list
        inside of a dictionary with an ``attributes`` key.

        **Example collection request**:

        .. sourcecode:: http

            POST /api/sites/1/attributes HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "attributes": [
                    {
                        "name": "owner",
                        "resource_name": "Network",
                    },
                    {
                        "name": "metro",
                        "resource_name": "Network",
                    }
                ]
            }

        **Example collection response**:

        Note that when creating a collection the response will not contain a
        ``Location`` header.

        .. sourcecode:: http

            HTTP/1.1 201 OK

            {
                "status": "ok",
                "data": {
                    "attributes": [
                        {
                            "id": 1,
                            "name": "owner",
                            "resource_name": "Network",
                            "required": false,
                            "display": false,
                            "multi": false,
                            "constraints": {
                                "allow_empty": false,
                                "pattern": "",
                                "valid_values": []
                            }
                        },
                        {
                            "id": 2,
                            "name": "metro",
                            "resource_name": "Network",
                            "required": false,
                            "display": false,
                            "multi": false,
                            "constraints": {
                                "allow_empty": false,
                                "pattern": "",
                                "valid_values": []
                            }
                        }
                    ],
                    "total": 2
                }
            }

        :permissions: * **admin**

        :param site_id: ID of the Site where this should be created.
        :type site_id: int

        :reqjson string name: The name of the Attribute
        :reqjson string resource_name: The type of resource this attribute is for (e.g. Network)
        :reqjson string description: (*optional*) A helpful description of
                                     the Attribute
        :reqjson bool required: (*optional*) Whether this attribute should be required.
        :reqjson bool display: (*optional*) Whether this attribute should be be displayed
                               by default in UIs.
        :reqjson bool multi: (*optional*) Whether the attribute should be treated as a
                             list type.

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.
        :reqheader X-NSoT-Email: required for all api requests.

        :resheader Location: URL to the created resource.

        :statuscode 201: The site was successfully created.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site at site_id was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        # Fetch collection from the body or the body itself
        objects = self.jbody.get('attributes', [self.jbody])

        attributes = []
        for obj in objects:
            try:
                name = obj["name"]
                description = obj.get("description")
                resource_name = obj["resource_name"]
                required = obj.get("required", False)
                display = obj.get("display", False)
                multi = obj.get("multi", False)
                constraints = obj.get("constraints", {})
            except KeyError as err:
                raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

            attribute = self.create_object(
                site_id, name, description, resource_name, required, display,
                multi, constraints, commit=False
            )
            attributes.append(attribute.to_dict())
        else:
            self.session.commit()

        # Return a collection w/ no Location header
        if len(attributes) > 1:
            self.created(
                data={"attributes": attributes, "total": len(attributes)}
            )
        # Or return a single object
        else:
            attribute = attributes[0]
            uri = "/api/sites/{}/attributes/{}".format(site_id, attribute['id'])
            self.created(uri, {"attribute": attribute})

    def create_object(self, site_id, name, description, resource_name, required,
                      display, multi, constraints, commit=True):
        """Create an Attribute object."""
        site_id = int(site_id)
        try:
            attribute = models.Attribute.create(
                self.session, self.current_user.id,
                site_id=site_id, name=name, description=description,
                resource_name=resource_name, required=required, display=display,
                multi=multi, constraints=constraints, commit=commit
            )
        except IntegrityError as err:
            raise exc.Conflict(str(err.orig))
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)
        return attribute

    def get(self, site_id):
        """ **Get all Attributes**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/attributes HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "attributes": [
                        {
                            "id": 1,
                            "site_id": 1,
                            "name": "vlan",
                            "description": "",
                            "resource_name": "Network",
                            "required": false,
                            "display": false,
                            "multi": false,
                            "constraints": {
                                "allow_empty": false,
                                "pattern": "",
                                "valid_values": []
                            }
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Attributes from.
        :type site_id: int

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.
        :query string name: (*optional*) Filter to attribute with name
        :query string resource_name: (*optional*) Filter to attributes for a specific resource.
        :query bool required: (*optional*) Filter to attributes that are required
        :query bool display: (*optional*) Filter to attributes meant to be displayed.
        :query bool multi: (*optional*) Filter on whether list type or not

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        name = self.get_argument("name", None)
        resource_name = self.get_argument("resource_name", None)
        required = self.get_argument("required", None)
        display = self.get_argument("display", None)
        multi = self.get_argument("multi", None)

        attributes = self.session.query(models.Attribute).filter_by(
            site_id=site_id
        )

        if name is not None:
            attributes = attributes.filter_by(name=name)
        if resource_name is not None:
            attributes = attributes.filter_by(resource_name=resource_name)
        if required is not None:
            attributes = attributes.filter_by(required=qpbool(required))
        if display is not None:
            attributes = attributes.filter_by(display=qpbool(display))
        if multi is not None:
            attributes = attributes.filter_by(multi=qpbool(multi))

        offset, limit = self.get_pagination_values()
        attributes, total = self.paginate_query(attributes, offset, limit)

        self.success({
            "attributes": [attribute.to_dict() for attribute in attributes],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class AttributeHandler(ApiHandler):
    def get(self, site_id, attribute_id):
        """ **Get a specific Attribute**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/attributes/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "attribute": {
                        "id": 1,
                        "site_id": 1,
                        "name": "vlan",
                        "description": "",
                        "resource_name": "Network",
                        "required": false,
                        "display": false,
                        "multi": false,
                        "constraints": {
                            "allow_empty": false,
                            "pattern": "",
                            "valid_values": []
                        }
                    }
                }
            }

        :param site_id: ID of the Site this Attribute is under.
        :type site_id: int

        :param attribute_id: ID of the Attribute being retrieved.
        :type attribute_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site or Attribute was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        attribute = self.session.query(models.Attribute).filter_by(
            id=attribute_id,
            site_id=site_id
        ).scalar()

        if not attribute:
            raise exc.NotFound(
                "No such Attribute found at (site_id, id) = ({}, {})".format(
                    site_id, attribute_id
                )
            )

        self.success({
            "attribute": attribute.to_dict(),
        })

    @any_perm("admin")
    def put(self, site_id, attribute_id):
        """ **Update an Attribute**

        **Example Request**:

        .. sourcecode:: http

            PUT /api/sites/1/attributes/1 HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "description": "Attribute Description",
                "required": true
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "attribute": {
                        "id": 1,
                        "name": "vlan",
                        "description": "Attribute Description",
                        "resource_name": "Network",
                        "required": true,
                        "display": false,
                        "multi": false,
                        "constraints": {
                            "allow_empty": false,
                            "pattern": "",
                            "valid_values": []
                        }
                    }
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param attribute_id: ID of the Attribute being updated.
        :type attribute_id: int

        :reqjson string description: (*optional*) A helpful description of
                                     the Attribute
        :reqjson bool required: (*optional*) Whether this attribute should be required.
        :reqjson bool display: (*optional*) Whether this attribute should be be displayed
                               by default in UIs.
        :reqjson bool multi: (*optional*) Whether the attribute should be treated as a
                             list type.

        :reqheader Content-Type: The server expects application/json.
        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Attribute was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        attribute = self.session.query(models.Attribute).filter_by(
            id=attribute_id,
            site_id=site_id
        ).scalar()

        if not attribute:
            raise exc.NotFound(
                "No such Attribute found at (site_id, id) = ({}, {})".format(
                    site_id, attribute_id
                )
            )

        try:
            description = self.jbody.get("description", "")
            required = self.jbody.get("required", False)
            display = self.jbody.get("display", False)
            multi = self.jbody.get("multi", False)
            constraints = self.jbody.get("constraints", {})
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

        try:
            attribute.update(
                self.current_user.id,
                description=description, required=required,
                display=display, multi=multi, constraints=constraints
            )
        except IntegrityError as err:
            raise exc.Conflict(str(err.orig))
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)

        self.success({
            "attribute": attribute.to_dict(),
        })

    @any_perm("admin")
    def delete(self, site_id, attribute_id):
        """ **Delete an Attribute**

        **Example Request**:

        .. sourcecode:: http

            DELETE /api/sites/1/attributes/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "message": Attribute 1 deleted."
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param attribute_id: ID of the Attribute being deleted.
        :type attribute_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Attribute was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        attribute = self.session.query(models.Attribute).filter_by(
            id=attribute_id,
            site_id=site_id
        ).scalar()

        if not attribute:
            raise exc.NotFound(
                "No such Attribute found at (site_id, id) = ({}, {})".format(
                    site_id, attribute_id
                )
            )

        try:
            attribute.delete(self.current_user.id)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)

        self.success({
            "message": "Attribute {} deleted from Site {}.".format(
                attribute_id, site_id
            ),
        })


class DevicesHandler(ApiHandler):

    @any_perm("admin")
    def post(self, site_id):
        """ **Create a Device or collection of Devices**

        **Example Request**:

        .. sourcecode:: http

            POST /api/sites/1/devices HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "hostname": "foobarhost",
                "attributes": {
                    "owner": "network-tools"
                }
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 201 OK
            Location: /api/sites/1/devices/1

            {
                "status": "ok",
                "data": {
                    "device": {
                        "id": 1,
                        "site_id": 1,
                        "hostname": "foobarhost",
                        "attributes": {"owner": "network-tools"}
                    }
                }
            }

        You may also create a collection of objects by submitting them as a list
        inside of a dictionary with an ``devices`` key.

        **Example collection request**:

        .. sourcecode:: http

            POST /api/sites/1/devices HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "devices": [
                    {"hostname": "foobarhost1"},
                    {"hostname": "foobarhost2"}
                ]
            }

        **Example collection response**:

        Note that when creating a collection the response will not contain a
        ``Location`` header.

        .. sourcecode:: http

            HTTP/1.1 201 OK

            {
                "status": "ok",
                "data": {
                    "devices": [
                        {
                            "id": 1,
                            "hostname": "foobarhost1",
                            "site_id": 1,
                            "attributes": {}
                        },
                        {
                            "id": 2,
                            "hostname": "foobarhost2",
                            "site_id": 1,
                            "attributes": {}
                        }
                    ],
                    "total": 2
                }
            }

        :permissions: * **admin**

        :param site_id: ID of the Site where this should be created.
        :type site_id: int

        :reqjson string hostname: The hostname of the network device.
        :reqjson object attributes: (*optional*) An object of key/value pairs
                                    attached to this device.

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.
        :reqheader X-NSoT-Email: required for all api requests.

        :resheader Location: URL to the created resource.

        :statuscode 201: The site was successfully created.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site at site_id was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        # Fetch collection from the body or the body itself as a list.
        objects = self.jbody.get('devices', [self.jbody])

        devices = []
        for obj in objects:
            try:
                hostname = obj["hostname"]
                attributes = obj.get("attributes", {})
            except KeyError as err:
                raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

            device = self.create_object(
                site_id, hostname, attributes, commit=False
            )
            devices.append(device.to_dict())
        else:
            self.session.commit()

        # Return a collection w/ no Location header
        if len(devices) > 1:
            self.created(data={"devices": devices, "total": len(devices)})
        # Or return a single object
        else:
            device = devices[0]
            uri = "/api/sites/{}/devices/{}".format(site_id, device['id'])
            self.created(uri, {"device": device})

    def create_object(self, site_id, hostname, attributes, commit=True):
        """Create a Device object."""
        site_id = int(site_id)
        try:
            device = models.Device.create(
                self.session, self.current_user.id, site_id=site_id,
                hostname=hostname, attributes=attributes, commit=commit
            )
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)
        except (ValueError, exc.ValidationError) as err:
            raise exc.BadRequest(err.message)
        return device

    def get(self, site_id):
        """ **Get all Devices**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/devices HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "devices": [
                        {
                            "id": 1,
                            "site_id": 1,
                            "hostname": "foobarhost",
                            "attributes": {"owner": "network-tools"}
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Devices from.
        :type site_id: int

        :query string hostname:
            (*optional*) Filter to device with hostname
        :query string attributes:
            (*optional*) Filter to device with matching attribute in k=v
            format. Can be supplied multiple times.
        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        devices = site.devices()

        hostname = self.get_argument("hostname", None)
        attributes = self.get_arguments("attributes", [])

        if hostname is not None:
            devices = devices.filter_by(hostname=hostname)

        # Just iterate the attributes and try to look them up as if they are k=v
        # and naively do an intersection query.
        for attribute in attributes:
            key, _, val = attribute.partition('=')
            next_set = site.devices(key, val)
            devices = devices.filter(
                models.Device.id.in_(
                    next_set.with_entities(models.Device.id)
                )
            )

        offset, limit = self.get_pagination_values()
        devices, total = self.paginate_query(devices, offset, limit)

        self.success({
            "devices": [device.to_dict() for device in devices],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class DevicesQueryHandler(ApiHandler):
    """
    Use set operations syntax to perform queries for Devices.
    """
    def get(self, site_id):
        """
        **Intersection request**:

        .. sourcecode:: http

            GET /api/sites/1/devices/query?query=foo%3Dbar HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Intersection response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "devices": [
                        {
                            "attributes": {
                                "owner": "team-networking",
                                "foo": "bar"
                            },
                            "hostname": "foo-bar1",
                            "site_id": 1,
                            "id": 1
                        },
                        {
                            "attributes": {
                                "owner": "jathan",
                                "foo": "bar"
                            },
                            "hostname": "foo-bar3",
                            "site_id": 1,
                            "id": 3
                        }
                    ]
                }
            }

        **Difference request**:

        .. sourcecode:: http

            GET /api/sites/1/devices/query?query=foo%3Dbar+-owner%3Djathan HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Difference response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "devices": [
                        {
                            "attributes": {
                                "owner": "team-networking",
                                "foo": "bar"
                            },
                            "hostname": "foo-bar1",
                            "site_id": 1,
                            "id": 1
                        }
                    ]
                }
            }

        :param site_id: ID of the Site to retrieve Devices from.
        :type site_id: int

        :param query: Set query representation
        :type query: str

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        query = self.get_argument("query", "")

        # Try to convert attributes into a dict.
        try:
            attributes = parse_set_query(query)
        except (ValueError, TypeError):
            attributes = []

        devices = site.devices()

        # Iterate a/v pairs and combine query results using MySQL-compatible set
        # operations w/ the ORM
        for action, name, value in attributes:
            next_set = site.devices(attribute_name=name, attribute_value=value)

            # This is the MySQL-compatible manual implementation
            if action == 'union':
                log.debug('SQL UNION')
                devices = devices.union(next_set)
            elif action == 'difference':
                log.debug('SQL DIFFERENCE')
                devices = devices.filter(
                    models.Device.id.notin_(
                        next_set.with_entities(models.Device.id)
                    )
                )
            elif action == 'intersection':
                log.debug('SQL INTERSECTION')
                devices = devices.filter(
                    models.Device.id.in_(
                        next_set.with_entities(models.Device.id)
                    )
                )
            else:
                raise exc.BadRequest('BAD SET QUERY: %r' % (action,))

        # Always order the objects by ID
        devices = devices.order_by(models.Device.id)

        offset, limit = self.get_pagination_values()
        devices, total = self.paginate_query(devices, offset, limit)

        self.success({
            "devices": [device.to_dict() for device in devices],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class DeviceHandler(ApiHandler):
    def get(self, site_id, device_id):
        """ **Get a specific Device**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/devices/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "device": {
                        "id": 1,
                        "site_id": 1,
                        "hostname": "foobarhost",
                        "attributes": {"owner": "network-tools"}
                    }
                }
            }

        :param site_id: ID of the Site this Device is under.
        :type site_id: int

        :param device_id: ID of the Device being retrieved.
        :type device_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site or Device was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        device = self.session.query(models.Device).filter_by(
            id=device_id,
            site_id=site_id
        ).scalar()

        if not device:
            raise exc.NotFound(
                "No such Device found at (site_id, id) = ({}, {})".format(
                    site_id, device_id
                )
            )

        self.success({
            "device": device.to_dict(),
        })

    @any_perm("admin")
    def put(self, site_id, device_id):
        """ **Update a Device**

        **Example Request**:

        .. sourcecode:: http

            PUT /api/sites/1/devices/1 HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "hostname": "foobarhost",
                "attributes": {"owner": "network-eng"}
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "device": {
                        "id": 1,
                        "site_id": 1,
                        "hostname": "foobarhost",
                        "attributes": {"owner": "network-eng"}
                    }
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param device_id: ID of the Attribute being updated.
        :type device_id: int

        :reqjson string hostname: The hostname of the device.
        :reqjson object attributes: (*optional*) A key/value pair of attributes
                                    attached to the device

        :reqheader Content-Type: The server expects application/json.
        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Device was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        device = self.session.query(models.Device).filter_by(
            id=device_id,
            site_id=site_id
        ).scalar()

        if not device:
            raise exc.NotFound(
                "No such Device found at (site_id, id) = ({}, {})".format(
                    site_id, device_id
                )
            )

        try:
            hostname = self.jbody["hostname"]
            attributes = self.jbody.get("attributes", {})
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

        try:
            device.update(self.current_user.id, hostname=hostname, attributes=attributes)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)

        self.success({
            "device": device.to_dict(),
        })

    @any_perm("admin")
    def delete(self, site_id, device_id):
        """ **Delete a Device**

        **Example Request**:

        .. sourcecode:: http

            DELETE /api/sites/1/devices/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "message": Device 1 deleted."
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param device_id: ID of the Device being deleted.
        :type device_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Attribute was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        device = self.session.query(models.Device).filter_by(
            id=device_id,
            site_id=site_id
        ).scalar()

        if not device:
            raise exc.NotFound(
                "No such Device found at (site_id, id) = ({}, {})".format(
                    site_id, device_id
                )
            )

        try:
            device.delete(self.current_user.id)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)

        self.success({
            "message": "Device {} deleted from Site {}.".format(
                device_id, site_id
            ),
        })


class NetworksHandler(ApiHandler):

    @any_perm("admin")
    def post(self, site_id):
        """ **Create a Network or collection of Networks**

        **Example Request**:

        .. sourcecode:: http

            POST /api/sites/1/networks HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "cidr": "10.0.0.0/8",
                "attributes": {
                    "vlan": "23"
                }
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 201 OK
            Location: /api/sites/1/networks/1

            {
                "status": "ok",
                "data": {
                    "network": {
                        "id": 1,
                        "parent_id": null,
                        "site_id": 1,
                        "is_ip": false,
                        "ip_version": "4",
                        "network_address": "10.0.0.0",
                        "prefix_length": "8",
                        "attributes": {"vlan": "23"}
                    }
                }
            }

        You may also create a collection of objects by submitting them as a list
        inside of a dictionary with an ``networks`` key.

        **Example collection request**:

        .. sourcecode:: http

            POST /api/sites/1/networks HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "networks": [
                    {"cidr": "10.0.0.0/8"},
                    {"cidr": "172.16.0.0/12"}
                ]
            }

        **Example collection response**:

        Note that when creating a collection the response will not contain a
        ``Location`` header.

        .. sourcecode:: http

            HTTP/1.1 201 OK

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "id": 1,
                            "parent_id": null,
                            "site_id": 1,
                            "is_ip": false,
                            "ip_version": "4",
                            "network_address": "10.0.0.0",
                            "prefix_length": "8",
                            "attributes": {}
                        },
                        {
                            "id": 2,
                            "parent_id": null,
                            "site_id": 1,
                            "is_ip": false,
                            "ip_version": "4",
                            "network_address": "172.16.0.0",
                            "prefix_length": "12",
                            "attributes": {}
                        }
                    ],
                    "total": 2
                }
            }

        :permissions: * **admin**

        :param site_id: ID of the Site where this should be created.
        :type site_id: int

        :reqjson string cidr: A network or ip address in CIDR notation.
        :reqjson object attributes: (*optional*) An object of key/value pairs
                                    attached to this network.

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.
        :reqheader X-NSoT-Email: required for all api requests.

        :resheader Location: URL to the created resource.

        :statuscode 201: The site was successfully created.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site at site_id was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        # Fetch collection from the body or the body itself
        objects = self.jbody.get('networks', [self.jbody])

        networks = []
        for obj in objects:
            try:
                cidr = obj["cidr"]
                attributes = obj.get("attributes", {})
            except KeyError as err:
                raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

            network = self.create_object(
                site_id, cidr, attributes, commit=False
            )
            networks.append(network.to_dict())
        else:
            self.session.commit()

        # Return a collection w/ no Location header
        if len(networks) > 1:
            self.created(data={"networks": networks, "total": len(networks)})
        # Or return a single object
        else:
            network = networks[0]
            uri = "/api/sites/{}/networks/{}".format(site_id, network['id'])
            self.created(uri, {"network": network})

    def create_object(self, site_id, cidr, attributes, commit=True):
        """Create a Network object."""
        site_id = int(site_id)
        try:
            network = models.Network.create(
                self.session, self.current_user.id, site_id,
                cidr=cidr, attributes=attributes,
            )
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)
        except (ValueError, exc.ValidationError) as err:
            raise exc.BadRequest(err.message)
        return network

    def get(self, site_id):
        """ **Get all Networks**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/networks HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "id": 1,
                            "parent_id": null,
                            "site_id": 1,
                            "is_ip": false,
                            "ip_version": "4",
                            "network_address": "10.0.0.0",
                            "prefix_length": "8",
                            "attributes": {}
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Networks from.
        :type site_id: int

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.
        :query bool root_only: (*optional*) Filter to root networks.
                               Default: false
        :query bool include_networks: (*optional*) Include non-IP networks.
                                      Default: true
        :query bool include_ips: (*optional*) Include IP addresses.
                                 Default: false

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        root_only = qpbool(self.get_argument("root_only", False))
        include_networks = qpbool(self.get_argument("include_networks", True))
        include_ips = qpbool(self.get_argument("include_ips", False))

        networks = site.networks(
            root=root_only, include_ips=include_ips,
            include_networks=include_networks
        )

        offset, limit = self.get_pagination_values()
        networks, total = self.paginate_query(networks, offset, limit)

        self.success({
            "networks": [network.to_dict() for network in networks],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class NetworksQueryHandler(ApiHandler):
    """
    Use set operations syntax to perform queries for Networks.
    """
    def get(self, site_id):
        """
        **Intersection request**:

        .. sourcecode:: http

            GET /api/sites/1/networks/query?query=owner%3Djathan HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Intersection response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "is_ip": false,
                            "site_id": 1,
                            "network_address": "10.0.0.0",
                            "parent_id": null,
                            "prefix_length": 8,
                            "ip_version": "4",
                            "attributes": {
                                "owner": "jathan",
                                "cluster": ""
                            },
                            "id": 1
                        },
                        {
                            "is_ip": false,
                            "site_id": 1,
                            "network_address": "10.20.30.0",
                            "parent_id": 1,
                            "prefix_length": 24,
                            "ip_version": "4",
                            "attributes": {
                                "owner": "jathan",
                                "foo": "bar"
                            },
                            "id": 4
                        }
                    ]
                }
            }

        **Difference request**:

        .. sourcecode:: http

            GET /api/sites/1/networks/query?query=owner%3Djathan+-foo%3Dbar HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Difference response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "is_ip": false,
                            "site_id": 1,
                            "network_address": "10.0.0.0",
                            "parent_id": null,
                            "prefix_length": 8,
                            "ip_version": "4",
                            "attributes": {
                                "owner": "jathan",
                                "cluster": ""
                            },
                            "id": 1
                        }
                    ]
                }
            }

        :param site_id: ID of the Site to retrieve Devices from.
        :type site_id: int

        :param query: Set query representation
        :typ query: str

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """

        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        query = self.get_argument("query", "")

        # Try to convert attributes into a dict.
        try:
            attributes = parse_set_query(query)
        except (ValueError, TypeError):
            attributes = []

        networks = site.networks()

        # Iterate a/v pairs and combine query results using MySQL-compatible set
        # operations w/ the ORM
        for action, name, value in attributes:
            next_set = site.networks(attribute_name=name,
                                     attribute_value=value)

            # This is the MySQL-compatible manual implementation
            if action == 'union':
                log.debug('SQL UNION')
                networks = networks.union(next_set)
            elif action == 'difference':
                log.debug('SQL DIFFERENCE')
                networks = networks.filter(
                    models.Network.id.notin_(
                        next_set.with_entities(models.Network.id)
                    )
                )
            elif action == 'intersection':
                log.debug('SQL INTERSECTION')
                networks = networks.filter(
                    models.Network.id.in_(
                        next_set.with_entities(models.Network.id)
                    )
                )
            else:
                raise exc.BadRequest('BAD SET QUERY: %r' % (action,))

        # Always order the objects by ID
        networks = networks.order_by(models.Network.id)

        offset, limit = self.get_pagination_values()
        networks, total = self.paginate_query(networks, offset, limit)

        self.success({
            "networks": [network.to_dict() for network in networks],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class NetworkHandler(ApiHandler):
    def get(self, site_id, network_id):
        """ **Get a specific Network**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/networks/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "network": {
                        "id": 1,
                        "parent_id": null,
                        "site_id": 1,
                        "is_ip": false,
                        "ip_version": "4",
                        "network_address": "10.0.0.0",
                        "prefix_length": "8",
                        "attributes": {}
                    }
                }
            }

        :param site_id: ID of the Site this Network is under.
        :type site_id: int

        :param network_id: ID of the Network being retrieved.
        :type network_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site or Network was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        network = self.session.query(models.Network).filter_by(
            id=network_id,
            site_id=site_id
        ).scalar()

        if not network:
            raise exc.NotFound(
                "No such Network found at (site_id, id) = ({}, {})".format(
                    site_id, network_id
                )
            )

        self.success({
            "network": network.to_dict(),
        })

    @any_perm("admin")
    def put(self, site_id, network_id):
        """ **Update a Network**

        **Example Request**:

        .. sourcecode:: http

            PUT /api/sites/1/networks/1 HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "attributes": {"vlan": "4"}
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "network": {
                        "id": 1,
                        "parent_id": null,
                        "site_id": 1,
                        "is_ip": false,
                        "ip_version": "4",
                        "network_address": "10.0.0.0",
                        "prefix_length": "8",
                        "attributes": {"vlan": "4"}
                    }
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param network_id: ID of the Attribute being updated.
        :type network_id: int

        :reqjson object attributes: (*optional*) A key/value pair of attributes
                                    attached to the network

        :reqheader Content-Type: The server expects application/json.
        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Network was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        network = self.session.query(models.Network).filter_by(
            id=network_id,
            site_id=site_id
        ).scalar()

        if not network:
            raise exc.NotFound(
                "No such Network found at (site_id, id) = ({}, {})".format(
                    site_id, network_id
                )
            )

        try:
            attributes = self.jbody.get("attributes", {})
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

        try:
            network.update(self.current_user.id, attributes=attributes)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)
        except exc.ValidationError as err:
            raise exc.BadRequest(err.message)

        self.success({
            "network": network.to_dict(),
        })

    @any_perm("admin")
    def delete(self, site_id, network_id):
        """ **Delete a Network**

        **Example Request**:

        .. sourcecode:: http

            DELETE /api/sites/1/networks/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "message": Network 1 deleted."
                }
            }


        :permissions: * **admin**

        :param site_id: ID of the Site that should be updated.
        :type site_id: int

        :param network_id: ID of the Network being deleted.
        :type network_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or Attribute was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        network = self.session.query(models.Network).filter_by(
            id=network_id,
            site_id=site_id
        ).scalar()

        if not network:
            raise exc.NotFound(
                "No such Network found at (site_id, id) = ({}, {})".format(
                    site_id, network_id
                )
            )

        try:
            network.delete(self.current_user.id)
        except IntegrityError as err:
            raise exc.Conflict(err.orig.message)

        self.success({
            "message": "Network {} deleted from Site {}.".format(
                network_id, site_id
            ),
        })


class NetworkSubnetsHandler(ApiHandler):
    def get(self, site_id, network_id):
        """ **Get subnets of a Network**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/networks/1/subnets HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "id": 2,
                            "parent_id": 1,
                            "site_id": 1,
                            "is_ip": false,
                            "ip_version": "4",
                            "network_address": "10.0.0.0",
                            "prefix_length": "24",
                            "attributes": {}
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Networks from.
        :type site_id: int

        :param network_id: ID of the Network we're requesting subnets from.
        :type network_id: int

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.
        :query bool direct: (*optional*) Return only direct subnets.
                            Default: false
        :query bool include_networks: (*optional*) Include non-IP networks.
                                      Default: true
        :query bool include_ips: (*optional*) Include IP addresses.
                                 Default: false

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site or Network was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        network = self.session.query(models.Network).filter_by(
            id=network_id,
            site_id=site_id
        ).scalar()

        if not network:
            raise exc.NotFound(
                "No such Network found at (site_id, id) = ({}, {})".format(
                    site_id, network_id
                )
            )

        direct = qpbool(self.get_argument("direct", False))
        include_networks = qpbool(self.get_argument("include_networks", True))
        include_ips = qpbool(self.get_argument("include_ips", False))

        networks = network.subnets(
            self.session, direct=direct,
            include_ips=include_ips, include_networks=include_networks
        )

        offset, limit = self.get_pagination_values()
        networks, total = self.paginate_query(networks, offset, limit)

        self.success({
            "networks": [network.to_dict() for network in networks],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class NetworkSupernetsHandler(ApiHandler):
    def get(self, site_id, network_id):
        """ **Get supernets of a Network**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/networks/2/supernets HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "networks": [
                        {
                            "id": 1,
                            "parent_id": null,
                            "site_id": 1,
                            "is_ip": false,
                            "ip_version": "4",
                            "network_address": "10.0.0.0",
                            "prefix_length": "8",
                            "attributes": {}
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Networks from.
        :type site_id: int

        :param network_id: ID of the Network we're requesting supernets from.
        :type network_id: int

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.
        :query bool direct: (*optional*) Return only direct supernets.
                            Default: false

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site or Network was not found.
        """
        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        network = self.session.query(models.Network).filter_by(
            id=network_id,
            site_id=site_id
        ).scalar()

        if not network:
            raise exc.NotFound(
                "No such Network found at (site_id, id) = ({}, {})".format(
                    site_id, network_id
                )
            )

        direct = qpbool(self.get_argument("direct", False))

        networks = network.supernets(self.session, direct=direct)

        offset, limit = self.get_pagination_values()
        networks, total = self.paginate_query(networks, offset, limit)

        self.success({
            "networks": [network.to_dict() for network in networks],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class ChangesHandler(ApiHandler):
    def get(self, site_id):
        """ **Get all Changes**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/changes HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "changes": [
                        {
                            "id": 1,
                            "site": {
                                "id": 1,
                                "name": "Site 1",
                                "description": ""
                            },
                            "user": {
                                "id": 1,
                                "email": "user@localhost"
                            },
                            "change_at": 1420062748,
                            "event": "Create",
                            "resource_name": "Site",
                            "resource_id": 1,
                            "resource": {
                                "name": "New Site",
                                "description": ""
                            },
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :param site_id: ID of the Site to retrieve Changes from.
        :type site_id: int

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.
        :query string event: (*optional*) Filter result to specific event.
                             Default: false
        :query string resource_name: (*optional*) Filter result to specific
                                     resource type. Default: false
        :query string resource_id: (*optional*) Filter result to specific resource id.
                                   Requires: resource_name, Default: false

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: Invalid query parameter values.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site at site_id was not found.
        """

        event = self.get_argument("event", None)
        if event is not None and event not in models.CHANGE_EVENTS:
            raise exc.BadRequest("Invalid event.")

        resource_name = self.get_argument("resource_name", None)
        if resource_name is not None and resource_name not in models.RESOURCE_BY_NAME:
            raise exc.BadRequest("Invalid resource type.")

        resource_id = self.get_argument("resource_id", None)
        if resource_id is not None and resource_name is None:
            raise exc.BadRequest("resource_id requires resource_name to be set.")

        changes = self.session.query(models.Change)

        if site_id is not None:
            site = self.session.query(models.Site).filter_by(id=site_id).scalar()
            if not site:
                raise exc.NotFound("No such Site found at id {}".format(site_id))
            changes = changes.filter_by(site_id=site_id)

        if event is not None:
            changes = changes.filter_by(event=event)

        if resource_name is not None:
            changes = changes.filter_by(resource_name=resource_name)

        if resource_id is not None:
            changes = changes.filter_by(resource_id=resource_id)

        changes = changes.order_by(desc("change_at"))

        offset, limit = self.get_pagination_values()
        changes, total = self.paginate_query(changes, offset, limit)

        self.success({
            "changes": [change.to_dict() for change in changes],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class ChangeHandler(ApiHandler):
    def get(self, site_id, change_id):
        """ **Get a specific Change**

        **Example Request**:

        .. sourcecode:: http

            GET /api/sites/1/changes/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "change": {
                        "id": 1,
                        "site": {
                            "id": 1,
                            "name": "Site 1",
                            "description": ""
                        },
                        "user": {
                            "id": 1,
                            "email": "user@localhost"
                        },
                        "change_at": 1420062748,
                        "event": "Create",
                        "resource_name": "Site",
                        "resource_id": 1,
                        "resource": {
                            "name": "New Site",
                            "description": ""
                        },
                    }
                }
            }

        :param site_id: ID of the Site to retrieve Changes from.
        :type site_id: int
        :param change_id: ID of the Change.
        :type change_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The Site/Change was not found.
        """

        change = self.session.query(models.Change).filter_by(
            id=change_id, site_id=site_id
        ).scalar()

        if not change:
            raise exc.NotFound(
                "No such Change ({}) at Site ({})".format(change_id, site_id)
            )

        self.success({
            "change": change.to_dict(),
        })


class AuthTokenLoginHandler(ApiHandler):
    """Used to authenticate users and return an auth_token."""

    def get_current_user(self):
        try:
            email = self.jbody['email']
            secret_key = self.jbody['secret_key']
        except KeyError as err:
            raise exc.Unauthorized('Missing required field: {}'.format(err.message))

        user = self.session.query(models.User).filter_by(email=email).scalar()

        # Make sure we've got a user and the secret_key is valid
        if user is not None and user.verify_secret_key(secret_key):
            return user  # Auth success

        raise exc.Unauthorized('Invalid email/secret_key')

    def post(self):
        """
        **Obtain an auth_token used for API calls.**

        **Example Request**:

        .. sourcecode:: http

            POST /api/authenticate HTTP/1.1
            Host: localhost
            Content-Type: application/json

            {
                "email": "user@localhost",
                "secret_key": "RmjJASeWQDHIoP7LMpSKGkofkoXnhbiBqauCnCR_InI=",
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "auth_token": "gAAAAAbu5=="
                }
            }

        :reqjson string email: User's email address
        :reqjson string secret_key: Secret key for the user

        :reqheader Content-Type: The server expects a json body specified with
                                 this header.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without valid credentials.
        """
        user = self.get_current_user()

        # Return the auth_token
        self.success({
            'auth_token': user.generate_auth_token()
        })


class AuthTokenVerifyHandler(ApiHandler):
    """Used to verify an auth_token."""
    def post(self):
        """
        **Verify an auth_token.**

        **Example Request**:

        .. sourcecode:: http

            POST /api/verify_token HTTP/1.1
            Host: localhost
            Content-Type: application/json
            Authorization: AuthToken user@localhost:gAAAAAbu5==

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": true
            }

        :reqheader Authorization: required for all api requests.
        :reqheader Content-Type: The server expects a json body specified with
                                 this header.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without valid credentials.
        """
        user = self.get_current_user()

        # Return 'ok'
        self.success(True)


class UsersHandler(ApiHandler):
    def get(self):
        """ **Get all Users**

        **Example Request**:

        .. sourcecode:: http

            GET /api/users HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "users": [
                        {
                            "id": 1
                            "email": "user@localhost"
                        }
                    ],
                    "limit": null,
                    "offset": 0,
                    "total": 1,
                }
            }

        :reqheader X-NSoT-Email: required for all api requests.

        :query int limit: (*optional*) Limit result to N resources.
        :query int offset: (*optional*) Skip the first N resources.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        """

        users = self.session.query(models.User)

        offset, limit = self.get_pagination_values()
        users, total = self.paginate_query(users, offset, limit)

        self.success({
            "users": [user.to_dict() for user in users],
            "limit": limit,
            "offset": offset,
            "total": total,
        })


class UserHandler(ApiHandler):
    def get(self, user_id):
        """ **Get a specific User**

        **Example Request**:

        .. sourcecode:: http

            GET /api/users/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "user": {
                        "id": 1
                        "email": "user@localhost"
                        "permissions": {
                            1: {
                                "user_id": 1,
                                "site_id": 1,
                                "permissions": ["admin"]
                            }
                        }
                    }
                }
            }

        :param user_id: ID of the User to retrieve or 0 for self.
        :type user_id: int

        :query bool with_secret_key: (*optional*) Include security_key
                                     (Self only.)

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The User was not found.
        """

        if user_id == "0":
            user = self.current_user
        else:
            user = self.session.query(models.User).filter_by(
                id=user_id,
            ).scalar()

        if not user:
            raise exc.NotFound(
                "No such User found at (id) = ({})".format(user_id)
            )

        user_kwargs = {
            "with_permissions": True,
        }

        with_secret_key = self.get_argument("with_secret_key", None)
        if with_secret_key is not None:
            if user != self.current_user:
                raise exc.Forbidden("Can't access secret_key of user that isn't you.")
            user_kwargs["with_secret_key"] = qpbool(with_secret_key)

        self.success({
            "user": user.to_dict(**user_kwargs),
        })


class UserRotateSecretKeyHandler(ApiHandler):
    def post(self, user_id):
        """ **Get a specific User**

        **Example Request**:

        .. sourcecode:: http

            POST /api/users/1/rotate_secret_key HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "secret_key": "k4ljk2345kl2j5k2fffml23",
                }
            }

        :param user_id: ID of the User to rotate key for or 0 for self.
        :type user_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The User was not found.
        """

        if user_id == "0":
            user = self.current_user
        else:
            user = self.session.query(models.User).filter_by(
                id=user_id,
            ).scalar()

        if user != self.current_user:
            raise exc.Forbidden("Can't access secret_key of user that isn't you.")

        if not user:
            raise exc.NotFound(
                "No such User found at (id) = ({})".format(user_id)
            )

        user.rotate_secret_key()
        self.session.commit()

        self.success({
            "secret_key": user.secret_key,
        })

class UserPermissionsHandler(ApiHandler):
    def get(self, user_id):
        """ **Get permissions for a specific User**

        **Example Request**:

        .. sourcecode:: http

            GET /api/users/1/permissions HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "permissions": {
                        1: {
                            "user_id": 1,
                            "site_id": 1,
                            "permissions": ["admin"]
                        }
                    }
                }
            }

        :param user_id: ID of the User to retrieve or 0 for self.
        :type user_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The User was not found.
        """

        if user_id == "0":
            user = self.current_user
        else:
            user = self.session.query(models.User).filter_by(
                id=user_id,
            ).scalar()

        if not user:
            raise exc.NotFound(
                "No such User found at (id) = ({})".format(user_id)
            )

        permissions = self.session.query(models.Permission).filter_by(
            user_id=user.id
        )

        self.success({
            "permissions": user.get_permissions(),
        })


class UserPermissionHandler(ApiHandler):
    def get(self, user_id, site_id):
        """ **Get permissions for a specific User and Site**

        **Example Request**:

        .. sourcecode:: http

            GET /api/users/1/permissions/1 HTTP/1.1
            Host: localhost
            X-NSoT-Email: user@localhost

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "permission": {
                        "user_id": 1,
                        "site_id": 1,
                        "permissions": ["admin"]
                    }
                }
            }

        :param user_id: ID of the User or 0 for self.
        :type user_id: int

        :param site_id: ID of the Site
        :type site_id: int

        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 401: The request was made without being logged in.
        :statuscode 404: The User or Site was not found.
        """

        if user_id == "0":
            user = self.current_user
        else:
            user = self.session.query(models.User).filter_by(
                id=user_id,
            ).scalar()
        if not user:
            raise exc.NotFound("No such User found at (id) = ({})".format(user_id))

        site = self.session.query(models.Site).filter_by(id=site_id).scalar()
        if not site:
            raise exc.NotFound("No such Site found at id {}".format(site_id))

        permission = self.session.query(models.Permission).filter_by(
            user_id=user.id, site_id=site_id
        ).scalar()

        if not permission:
            raise exc.NotFound(
                "No such Permission found at (user_id, site_id) = ({})".format(
                    user.id, site_id
                )
            )

        self.success({
            "permission": permission.to_dict(),
        })

    @any_perm("admin")
    def put(self, user_id, site_id):
        """ **Create/Update a Users Permissions for a Site**

        **Example Request**:

        .. sourcecode:: http

            PUT /api/users/2/permissions/1 HTTP/1.1
            Host: localhost
            Content-Type: application/json
            X-NSoT-Email: user@localhost

            {
                "permissions": ["networks"]
            }

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Content-Type: application/json

            {
                "status": "ok",
                "data": {
                    "permission": {
                        "user_id": 2,
                        "site_id": 1,
                        "permissions": ["networks"]
                    }
                }
            }

        :permissions: * **admin**

        :param user_id: ID of the User
        :type user_id: int

        :param site_id: ID of the Site
        :type site_id: int

        :reqjson array permissions: A list of permissions the user should have.

        :reqheader Content-Type: The server expects application/json.
        :reqheader X-NSoT-Email: required for all api requests.

        :statuscode 200: The request was successful.
        :statuscode 400: The request was malformed.
        :statuscode 401: The request was made without being logged in.
        :statuscode 403: The request was made with insufficient permissions.
        :statuscode 404: The Site or User was not found.
        :statuscode 409: There was a conflict with another resource.
        """
        permission = self.session.query(models.Permission).filter_by(
                user_id=user_id, site_id=site_id
        ).scalar()

        # If the permission exists it's safe to assume the user/site exists. If not
        # we're adding a new permission so verify that the user/site are valid.
        if not permission:
            user = self.session.query(models.User).filter_by(id=user_id).scalar()
            if not user:
                raise exc.NotFound("No such User found at (id) = ({})".format(user_id))

            site = self.session.query(models.Site).filter_by(id=site_id).scalar()
            if not site:
                raise exc.NotFound("No such Site found at id {}".format(site_id))

        try:
            permissions = self.jbody["permissions"]
        except KeyError as err:
            raise exc.BadRequest("Missing Required Argument: {}".format(err.message))

        try:
            if not permission:
                permission = models.Permission.create(
                    self.session, self.current_user.id,
                    user_id=user.id, site_id=site_id,
                    permissions=permissions
                )
            else:
                permission.update(
                    self.current_user.id,
                    permissions=permissions
                )
        except IntegrityError as err:
            raise exc.Conflict(str(err.orig))

        self.success({
            "permission": permission.to_dict(),
        })


class NotFoundHandler(ApiHandler):
    def get(self):
        raise exc.NotFound("Endpoint not found")



#!/usr/bin/env python

import os

import setuptools
from distutils.core import setup

execfile('nsot/version.py')

with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()

package_data = {}
def get_package_data(package, base_dir):
    for dirpath, dirnames, filenames in os.walk(base_dir):
        dirpath = dirpath[len(package)+1:]  # Strip package dir
        for filename in filenames:
            package_data.setdefault(package, []).append(os.path.join(dirpath, filename))
        for dirname in dirnames:
            get_package_data(package, dirname)

get_package_data("nsot", "nsot/static")
get_package_data("nsot", "nsot/templates")
get_package_data("nsot", "nsot/migrations")

kwargs = {
    "name": "nsot",
    "version": str(__version__),
    "packages": ["nsot", "nsot.handlers"],
    "package_data": package_data,
    "scripts": ["bin/nsot-server", "bin/nsot-ctl"],
    "description": "Network Source of Truth (IP Address Management).",
    "author": "Gary M. Josack",
    "maintainer": "Gary M. Josack",
    "author_email": "gary@dropbox.com",
    "maintainer_email": "gary@dropbox.com",
    "license": "Apache",
    "install_requires": required,
    "url": "https://github.com/dropbox/nsot",
    "download_url": "https://github.com/dropbox/nsot/archive/master.tar.gz",
    "classifiers": [
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
}

setup(**kwargs)



import pytest
from sqlalchemy.exc import IntegrityError

from nsot import exc
from nsot import models

from .fixtures import session, user, admin, site


def test_permission_creation(session, user, site):
    models.Permission.create(
        session, user.id,
        user_id=user.id, site_id=site.id,
        permissions=[]
    )

    permission = session.query(models.Permission).filter_by(
        site_id=site.id, user_id=user.id
    ).scalar()

    assert permission.user_id == user.id
    assert permission.site_id == site.id
    assert permission.permissions == []


def test_permission_conflict(session, user, site):
    models.Permission.create(
        session, user.id,
        user_id=user.id, site_id=site.id,
        permissions=[]
    )

    with pytest.raises(IntegrityError):
        models.Permission.create(
            session, user.id,
            user_id=user.id, site_id=site.id,
            permissions=["admin"]
        )

def test_permission_update(session, user, site):
    perm = models.Permission.create(
        session, user.id,
        user_id=user.id, site_id=site.id,
        permissions=[]
    )

    perm.update(user.id, permissions=["admin"])

    assert perm.permissions == ["admin"]



"""
Provide a ScopedSession object that is only imported as needed in
certain scopes (such as with classmethods).
"""

from sqlalchemy.orm import scoped_session
from .models import Session

ScopedSession = scoped_session(Session)



import pytest

from nsot import models


@pytest.fixture
def session(request, tmpdir):
    db_path = tmpdir.join("nsot.sqlite")
    db_engine = models.get_db_engine("sqlite:///%s" % db_path)

    models.Model.metadata.drop_all(db_engine)
    models.Model.metadata.create_all(db_engine)
    models.Session.configure(bind=db_engine)
    session = models.Session()

    def fin():
        session.close()
    request.addfinalizer(fin)

    return session


@pytest.fixture
def user(session):
    user = models.User(email="gary@localhost").add(session)
    session.commit()
    return user


@pytest.fixture
def admin(session):
    user = models.User(email="admin@localhost").add(session)
    session.commit()
    return user


@pytest.fixture
def site(session, admin):
    site = models.Site.create(
        session, admin.id,
        name="Test Site",
        description="This is a Test Site."
    )
    session.commit()
    return site
